@@ -1,14 +1,18 @@
 """Everything regarding parsing, generating and writing HTML belongs in here."""
 
 import collections
+import enum
 import html.parser
 import os
 import posixpath
+import re
 
 
+from . import document
+
 
 class ParseException(Exception):
-    """Exception to propagate an parsing error."""
+    """Exception to propagate a parsing error."""
     def __init__(self, msg, pos):
         self.msg = msg
         self.pos = pos
@@ -17,30 +21,48 @@ class ParseException(Exception):
     def __str__(self):
         return 'line {0.pos[0]}, {0.pos[1]}: {0.msg}'.format(self)
 
+def get_position(document, index):
+    """This returns the line number and position on line for the given String.
+    Note: lines and positions are counted from 0."""
+    line = document[:index+1].count('\n')
+    if document[index] == '\n':
+        return (line, 0)
+    newline = document[:index+1].rfind('\n')
+    newline = (newline if newline >= 0 else 0)
+    return (line, len(document[newline:index]))
 
-class EqnParser(html.parser.HTMLParser):
-    """This HTML parser parses a given document and tries to preserve the
-    original document as much as possible. It is saved in chunks which
-    reconstruct the whole document, when joined. An exception are formulas from
-    the <eq /> equation tag. Instead of saving those into a chunk too, a list is
-    added, where the first element is the position the formula was encountered,
-    the second is whether is displaymath 8set to True in this case) and the
-    third is the actual text chunk. so for getting all equations, one would do:
 
-            eqns = [e for e in parser.get_data()   if isinstance(e, list)]
+def find_anycase(where, what):
+    """Find with both lower or upper case."""
+    lower = where.find(what.lower())
+    upper = where.find(what.upper())
+    if lower >= 0:
+        return lower
+    else:
+        return upper
+
+class EqnParser:
+    """This parser parses <eq>...</eq> our of a document. It's not an HTML
+    parser, because the content within <eq>.*<eq> is parsed verbatim.
+    It also parses comments, to not consider formulas within comments. All other
+    cases are unhandled. Especially CData is problematic, although it seems like
+    a rare use case."""
+    class State(enum.Enum): # ([\s\S]*?) also matches newlines
+        Comment = re.compile(r'<!--([\s\S]*?)-->', re.MULTILINE)
+        Equation = re.compile(r'<\s*(?:eq|EQ)\s*(.*?)?>([\s\S.]+?)<\s*/\s*(?:eq|EQ)>',
+                re.MULTILINE)
+
+    HTML_ENTITY = re.compile(r'(&(:?#\d+|[a-zA-Z]+);)')
 
-    NOTE: The document is slightly altered. Tags will be lower case and some
-    spacing can be lost. The parser tries to preserve as much as possible, but
-    e.g. stand-alone tags like `<br />` would loose the space."""
     def __init__(self):
-        super().__init__(convert_charrefs=False)
+        self.__document = None
         self.__data = []
-        self.__lastchunk = []
         self.__encoding = None
-        self.in_eqn = False
 
     def feed(self, document):
-        """Overwrite to run a final step after parsing was completed."""
+        """Feed a string or a bytes instance. If a bytes instance is fed, an
+        encoding header has to be present, so that the encoding can be
+        extracted."""
         if isinstance(document, bytes): # try to guess encoding
             encoding = "UTF-8"
             if b'charset=' in document:
@@ -50,96 +72,114 @@ class EqnParser(html.parser.HTMLParser):
                     encoding = document[start:start+end].decode("utf-8")
             document = document.decode(encoding)
             self.__encoding = encoding
-        super().feed(document)
-        if self.in_eqn:
-            if isinstance(self.__data[-1], list):
-                start_pos = self.__data[-1][0]
-                raise ParseException("Unclosed equation environment.", start_pos)
-        self.__data += self.__lastchunk
-        self.__lastchunk = []
+        self.__document = document[:]
+        self._parse()
+
+    def find_with_offset(self, doc, start, what):
+        """This find method searches in the document for a given string, staking
+        the offset into account. REturned is the absolute position (so offset +
+        relative match position) or -1 for no hit."""
+        if isinstance(what, str):
+            pos = doc[start:].find(what)
+        else:
+            match = what.search(doc[start:])
+            pos = (-1 if not match else match.span()[0])
+        return (pos if pos == -1 else pos + start)
+
+
+    def _parse(self):
+        """This function parses the document, while maintaining state using the
+        State enum."""
+        in_document = lambda x: (False if x == -1 else True)
+        # maintain a lower-case copy, which eases searching, but doesn't affect
+        # the handler methods
+        doc = self.__document[:].lower()
+
+        end = len(self.__document) - 1
+        eq_start = re.compile(r'<\s*eq\s*(.*?)>')
+
+        start_pos = 0
+        while start_pos < end:
+            comment = self.find_with_offset(doc, start_pos, '<!--')
+            formula = self.find_with_offset(doc, start_pos, eq_start)
+            if in_document(comment) and in_document(formula): # both present, take closest
+                if comment < formula:
+                    self.__data.append(self.__document[start_pos:comment])
+                    start_pos = self.handle_comment(comment)
+                else:
+                    self.__data.append(self.__document[start_pos:formula])
+                    start_pos = self.handle_equation(formula)
+            elif in_document(formula):
+                self.__data.append(self.__document[start_pos:formula])
+                start_pos = self.handle_equation(formula)
+            elif in_document(comment):
+                self.__data.append(self.__document[start_pos:comment])
+                start_pos = self.handle_comment(comment)
+            else: # only data left
+                self.__data.append(self.__document[start_pos:])
+                start_pos = end
+
+
+    def handle_equation(self, start_pos):
+        """Parse an equation. The given offset should mark the beginning of this
+        equation."""
+        # get line and column of `start_pos`
+        lnum, pos = get_position(self.__document, start_pos)
+
+        match = EqnParser.State.Equation.value.search(self.__document[start_pos:])
+        if not match:
+            next_eq = find_anycase(self.__document[start_pos+1:], '<eq')
+            closing = find_anycase(self.__document[start_pos:], '</eq>')
+            if next_eq > -1 and closing > -1 and next_eq < closing:
+                raise ParseException("Unclosed tag found", (lnum, pos))
+            else:
+                raise ParseException("Malformed equation tag found", (lnum, pos))
+        end = start_pos + match.span()[1]
+        attrs, formula = match.groups()
+        if '<eq>' in formula or '<EQ' in formula:
+            raise ParseException("Invalid nesting of formulas detected.", (lnum,
+                pos))
+
+        # replace HTML entities
+        entity = EqnParser.HTML_ENTITY.search(formula)
+        while entity:
+            formula = re.sub(EqnParser.HTML_ENTITY,
+                    html.unescape(entity.groups()[0]), formula)
+            entity = EqnParser.HTML_ENTITY.search(formula)
+        attrs = attrs.lower()
+        displaymath = (True if attrs and 'env' in attrs and 'displaymath' in attrs
+                else False)
+        self.__data.append(((lnum, pos), # let line number count from 0 as well
+                displaymath, formula))
+        return end
+
+
+    def handle_comment(self, start_pos):
+        match = EqnParser.State.Comment.value.search(self.__document[start_pos:])
+        if not match:
+            lnum, pos = get_position(self.__document, start_pos)
+            # this could be a parser issue, too
+            raise ParseException("Improperly formatted comment found", (lnum,
+                pos))
+        self.__data.append('<!--%s-->' % match.groups()[0])
+        return start_pos + match.span()[1] # return end of match
 
     def get_encoding(self):
-        """Return encoding (if parsed) or none if no encoding was used (i.e. a
-        string was passed in."""
+        """Return the parsed encoding from the HTML meta data. If none was set,
+        UTF-8 is assumed."""
         return self.__encoding
 
-    def handle_starttag(self, tag, attrs):
-        if not tag == 'eq':
-            self.__lastchunk.append(self.get_starttag_text())
-        else:
-            if self.in_eqn:
-                raise ParseException(("Opening eq tag encountered while the "
-                    "last one wasn't yet closed."), self.getpos())
-            attrs = dict((k.lower(), v.lower()) for k,v in attrs)
-            displaymath = (True if 'env' in attrs and attrs['env'] == 'displaymath'
-                    else False)
-            # add already parsed elements:
-            self.__data += self.__lastchunk
-            self.__lastchunk = []
-            # initialize list item in self.__data for equation
-            lnum, pos = self.getpos() # position in document
-            self.__data.append([(lnum-1, pos), # let line number count from 0 as well
-                displaymath, None])
-            self.in_eqn = True
-
-    def handle_startendtag(self, tag, attrs):
-        if attrs:
-            self.__lastchunk.append('<{} {} />'.format(tag,
-                ' '.join(['%s="%s"' % (x[0], x[1]) for x in attrs])))
-        else:
-            self.__lastchunk.append('<%s />' % tag)
-
-    def handle_endtag(self, tag):
-        if self.in_eqn:
-            self.in_eqn = False
-            # add last chunk(s) to already saved eqn position in self.__data
-            if isinstance(self.__lastchunk, list):
-                self.__data[-1][2] = ''.join(self.__lastchunk)
-            else:
-                self.__data[-1][2] = self.__lastchunk
-            self.__lastchunk = [] # clear last chunk  for further processing
-        else:
-            self.__lastchunk.append('</%s>' % tag)
-
-    def handle_data(self, data):
-        if '\r' in data:
-            data = data.replace('\r', '')
-        self.__lastchunk.append(data)
-
-    def handle_entityref(self, name):
-        self.__lastchunk.append('&%s;' % name)
-
-    def handle_charref(self, name):
-        self.__lastchunk.append('&#%s;' % name)
-
-    def handle_comment(self, blah):
-        self.__lastchunk.append('<!--%s-->' % blah)
-
-    def handle_pi(self, instruction):
-        """Handle processing instruction."""
-        self.__lastchunk.append('<? %s>' % instruction)
-
-    def handle_unknown_decl(self, declaration):
-        self.__lastchunk.append('<!%s>' % declaration)
-
     def get_data(self):
-        """Return the chunks parsed from the HTML file. Everything which is not
-        a formula, formulas are lists. A formula looks like this:
-            (pos, display_math, formula)
-        pos: tuple of line number, position in line (both counting from 0)
-        display_math: boolean indicating whether formula is in display math
-        formula: the formula as a string.
-        """
-        return self.__data[:]
-
-    def error(self, message):
-        raise ParseException(message, ('unknown', 'unknown'))
+        """Return parsed chunks. These are either strings or tuples with formula
+        information, see class documentation."""
+        return list(x for x in self.__data if x) # filter empty bits
 
 
 def gen_id(formula):
-    """Generate an id for identifying a formula.
-    It will be valid to be used within a HTML attribute and it won't be too
-    long. If you happen to have a lot of formulas > 150 characters with exactly
+    """Generate an id for identifying a formula as an anchor in a document.
+    The generated ID is guaranteed to be valid in an XML attribute and it won't
+    exceed a certain length.
+    If you happen to have a lot of formulas > 150 characters with exactly
     the same content in the document, that'll cause a clash of id's."""
     # for some characters we just use a simple replacement (otherwise the
     # would be lost)
@@ -271,6 +311,8 @@ class HtmlImageFormatter: # ToDo: localisation
     nothing will be excluded."""
 
     EXCLUSION_FILE_NAME = 'outsourced-descriptions.html'
+    FORMATTING_COMMANDS = ['\\ ', '\\,', '\\;', '\\big', '\\Big', '\\left',
+            '\\right', '\\limits']
     HTML_TEMPLATE_HEAD = ('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"' +
         '\n  "http://www.w3.org/TR/html4/strict.dtd">\n<html>\n<head>\n' +
         '<meta http-equiv="content-type" content="text/html; charset=utf-8"/>' +
@@ -361,6 +403,34 @@ class HtmlImageFormatter: # ToDo: localisation
                     for formula in self.__cached_formula_pars.values()]))
             f.write('\n</body>\n</html>\n')
 
+    def postprocess_formula(self, formula):
+        """Formulas should not contain unicode characters or formatting
+        instructions, if they are part of the alt attribute. This distracts the
+        reader and stretches the formula. This method replaces non-ascii
+        characters, if requested and strips a few formatting commands."""
+        if self.__replace_nonascii:
+            formula = document.escape_unicode_in_formulas(formula,
+                    replace_alphabeticals=False)
+        # replace formatting-only symbols which distract the reader
+        formula_changed = True
+        while formula_changed:
+            formula_changed = False
+            for command in HtmlImageFormatter.FORMATTING_COMMANDS:
+                idx = formula.find(command)
+                # only replace if it's not after a \\ and not part of a longer command
+                if (idx > 0 and formula[idx-1] != '\\') or idx == 0:
+                    end = idx + len(command)
+                    # following conditions for replacement must be met:
+                    # command doesn't end on alphabet. char. and is followed by
+                    # same category; end of string reached or command does not
+                    # end on alphabet. char. at all
+                    if end >= len(formula) or not command[-1].isalpha() \
+                            or not formula[end].isalpha():
+                        formula = formula[:idx] + ' ' + formula[idx + len(command):]
+                        formula = formula.replace('  ', ' ')
+                        formula_changed = True
+        return formula
+
     def get_html_img(self, pos, formula, img_path, displaymath=False):
         """:param pos dictionary containing keys depth, height and width
         :param formula LaTeX alternative text
@@ -410,6 +480,7 @@ class HtmlImageFormatter: # ToDo: localisation
         :param displaymath whether or not formula is in display math (default: no)
         :returns string with formatted HTML image which also links to excluded
         formula"""
+        formula = self.postprocess_formula(formula)
         if self.__exclude_descriptions and \
                 len(formula) > self.__inline_maxlength:
             return self.format_excluded(pos, formula, img_path, displaymath)
